/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2011 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisSend.c - NDIS driver send functions.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver with code from
                            Microsoft sample driver.
   ---------------------------------------------------------------------------
*/


#include "target.h"
#include "NdisDevice.h"


/* -------------------------------------------------------------------------- */

/*
    GetFirstBuffer

Description:
    This function retrieves the first NDIS buffer in the NDIS packet.

Parameters:
    PMDL *ppMdl;
        Pointer to NDIS MDL pointer.

    ULONG Offset;
        Data offset in MDL.

    PULONG pulLength
        Buffer to store the length of the data buffer.

Return (PUCHAR):
     Pointer to data buffer in the NDIS net buffer.
*/

__inline
PUCHAR GetFirstBuffer (
    PMDL  *ppMdl,
    ULONG  Offset,
    PULONG pulLength )
{
    PUCHAR pBuffer;

    NdisQueryMdl(*ppMdl, &pBuffer, pulLength, NormalPagePriority);
    if (pBuffer && Offset)
    {
        pBuffer += Offset;
        *pulLength -= Offset;
    }

    return pBuffer;
}  // GetFirstBuffer


/*
    GetNextBuffer

Description:
    This function retrieves the next NDIS buffer in the NDIS packet.

Parameters:
    PMDL *ppMdl;
        Pointer to NDIS MDL pointer.

    PULONG pulLength
        Buffer to store the length of the data buffer.

Return (PUCHAR):
    Pointer to data buffer in the NDIS net buffer.
*/

__inline
PUCHAR GetNextBuffer (
    PMDL   *ppMdl,
    PULONG pulLength )
{
    PUCHAR pBuffer;

    *pulLength = 0;

    NdisGetNextMdl(*ppMdl, ppMdl);
    if (!*ppMdl)
    {
        return NULL;
    }
    NdisQueryMdl(*ppMdl, &pBuffer, pulLength, NormalPagePriority);
    return pBuffer;
}  // GetNextBuffer


static
VOID QueryPacket (
    IN  PNDIS_ADAPTER pAdapter,
    IN  PNET_BUFFER   NetBuffer,
    OUT PULONG        pBufferCount,
    OUT PULONG        pPacketLength )
{
    PMDL pMdl;

    // Address of current source buffer
    PUCHAR pBuffer;

    // Length of current source buffer
    ULONG BufferLength;
    ULONG BufferCount;
    ULONG DataLength;
    ULONG Length;
    ULONG MdlOffset;
    ULONG Offset;
    ULONG HdrLen = 0;
    BOOLEAN next = FALSE;
    BOOLEAN fFragment = FALSE;
    BOOLEAN fTcp = FALSE;
    BOOLEAN fUdp = FALSE;

    pMdl = NET_BUFFER_CURRENT_MDL(NetBuffer);
    MdlOffset = NET_BUFFER_CURRENT_MDL_OFFSET(NetBuffer);
    Offset = NET_BUFFER_DATA_OFFSET(NetBuffer);
    Length = NET_BUFFER_DATA_LENGTH(NetBuffer);

    DataLength = 0;
    BufferCount = 0;

    while (pMdl)
    {
        NdisQueryMdl(pMdl, &pBuffer, &BufferLength, NormalPagePriority);
        if (!pBuffer)
            break;
        if (!DataLength)
        {
            PUCHAR pData = pBuffer + MdlOffset;

            if (ETH_IS_BROADCAST(pData))
                FRAME_TYPE_FROM_SEND_NB(NetBuffer) =
                    NDIS_PACKET_TYPE_BROADCAST;
            else if (ETH_IS_MULTICAST(pData))
                FRAME_TYPE_FROM_SEND_NB(NetBuffer) =
                    NDIS_PACKET_TYPE_MULTICAST;
            else
                FRAME_TYPE_FROM_SEND_NB(NetBuffer) =
                    NDIS_PACKET_TYPE_DIRECTED;
            HdrLen = BufferLength - MdlOffset;
            if (pAdapter->ulUdpTestLimit && HdrLen >= 24)
            {
                PUSHORT pwData = (PUSHORT) &pData[12];
                UINT vlan_hlen = 0;
                UCHAR protocol;
               
                if (0x0081 == *pwData)
                {
                    vlan_hlen = 4;
                    pwData += 2;
                }
                if (0x0008 == *pwData)
                {
                    if (pwData[4] & 0xFF3F)
                        fFragment = TRUE;
                    protocol = pData[14 + 9 + vlan_hlen];
                    if (17 == protocol)
                        fUdp = TRUE;
                    else if (6 == protocol)
                        fTcp = TRUE;
                }
                else if (0xDD86 == *pwData)
                {
                    protocol = pData[14 + 6 + vlan_hlen];
                    if (0 == protocol)
                    {
                        if (HdrLen > 14 + 8 + vlan_hlen + 32)
                        {
                            protocol = pData[14 + 8 + vlan_hlen + 32];
                        }
                        else
                        {
                            HdrLen -= 14 + 8 + vlan_hlen + 32;
                            next = TRUE;
                        }
                    }
                    if (!next)
                    {
                        if (17 == protocol)
                            fUdp = TRUE;
                        else if (6 == protocol)
                            fTcp = TRUE;
                        else if (44 == protocol)
                            fFragment = TRUE;
                    }
                }
            }
        }
        else if (next) {
            UCHAR protocol = pBuffer[HdrLen];

            if (17 == protocol)
                fUdp = TRUE;
            else if (6 == protocol)
                fTcp = TRUE;
            else if (44 == protocol)
                fFragment = TRUE;
            next = FALSE;
        }
        BufferLength -= MdlOffset;
        MdlOffset = 0;
        DataLength += BufferLength;
        ++BufferCount;
        NdisGetNextMdl(pMdl, &pMdl);
    }
    if (fFragment && !fTcp)
        BufferCount = 1;
    *pBufferCount = BufferCount;
    *pPacketLength = DataLength;
}  // QueryPacket


#define MIN_PACKET_SIZE    40

/*
    CopyNetBuffer

Description:
    This function copies the packet down to the adapter buffer.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER NetBuffer
        Pointer to NDIS net buffer.

Return (ULONG):
    The length of the buffer.
*/

ULONG CopyNetBuffer (
    IN  PNDIS_ADAPTER pAdapter,
    IN  PNET_BUFFER   NetBuffer )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    PDMA_BUFFER pDma = pAdapter->TxBufInfo.pCurrent;
    PMDL pMdl;

    // Address of current source buffer
    PUCHAR pBuffer;

    // Length of current source buffer
    ULONG BufferLength;
    ULONG DataLength;
    ULONG Length;
    ULONG MdlOffset;
    ULONG Offset;

    pMdl = NET_BUFFER_CURRENT_MDL(NetBuffer);
    MdlOffset = NET_BUFFER_CURRENT_MDL_OFFSET(NetBuffer);
    Offset = NET_BUFFER_DATA_OFFSET(NetBuffer);
    Length = NET_BUFFER_DATA_LENGTH(NetBuffer);

    DataLength = 0;

    pBuffer = GetFirstBuffer(&pMdl, MdlOffset, &BufferLength);

    if (pBuffer)
    {
        HardwareSetTransmitBuffer(pHardware, pDma->sm.ulPhysical);
        if (Length < MIN_PACKET_SIZE)
            Length = MIN_PACKET_SIZE;
        HardwareSetTransmitLength(pHardware, Length);
    }

    while (pBuffer && Length > 0)
    {
        if (BufferLength > Length)
            BufferLength = Length;
        NdisMoveMemory(&pDma->sm.pVirtual[DataLength], pBuffer,
            BufferLength);
        DataLength += BufferLength;
        Length -= BufferLength;

        pBuffer = GetNextBuffer(&pMdl, &BufferLength);
    }

    if (DataLength && DataLength < MIN_PACKET_SIZE)
    {
        NdisZeroMemory(&pDma->sm.pVirtual[DataLength],
            MIN_PACKET_SIZE - DataLength);
    }

    NdisAdjustMdlLength(pDma->Mdl, DataLength);

    /* Flush buffer. */
    NdisFlushBuffer(pDma->Mdl, TRUE);

    return DataLength;
}  // CopyNetBuffer

/* -------------------------------------------------------------------------- */

#if 0
__inline
VOID HardwareChecksum (
    PNDIS_PACKET pPacket,
    PHARDWARE    pHardware )
{
    if ( NDIS_PROTOCOL_ID_TCP_IP == NDIS_GET_PACKET_PROTOCOL_TYPE( pPacket ) )
    {
        PNDIS_TCP_IP_CHECKSUM_PACKET_INFO pChecksum =
            ( PNDIS_TCP_IP_CHECKSUM_PACKET_INFO ) &
            NDIS_PER_PACKET_INFO_FROM_PACKET( pPacket,
                TcpIpChecksumPacketInfo );
        if ( pChecksum->Transmit.NdisPacketChecksumV4 )
        {
            if ( pChecksum->Transmit.NdisPacketTcpChecksum )
            {
                pHardware->m_TxDescInfo.pCurrent->sw.BufSize.tx.fCsumGenTCP =
                    TRUE;
            }
            if ( pChecksum->Transmit.NdisPacketUdpChecksum )
            {
                pHardware->m_TxDescInfo.pCurrent->sw.BufSize.tx.fCsumGenUDP =
                    TRUE;
            }
            if ( pChecksum->Transmit.NdisPacketIpChecksum )
            {
                pHardware->m_TxDescInfo.pCurrent->sw.BufSize.tx.fCsumGenIP =
                    TRUE;
            }
        }
    }
}  /* HardwareChecksum */
#endif


#define DRV_PACKET_ARRAY_SIZE  32

static
void ReturnTcb (
        PTCB_INFO pInfo,
    IN  PTCB      pTcb )
{
    if (pTcb->pSGL)
    {
        /* Ask network subsystem to free the mapping. */
        NdisMFreeNetBufferSGList(pInfo->hDma, pTcb->pSGL, pTcb->NetBuffer);
        pTcb->pSGL = NULL;
    }
    pTcb->NetBuffer = NULL;
    NdisInterlockedPushEntrySList(&pInfo->listHead, &pTcb->listEntry,
        &pInfo->listLock);
}  // ReturnTcb


/*
    TXNblReference

Description:
    This function adds a reference on a NBL that is being transmitted.
    The NBL won't be returned to the protocol until the last reference is
    released.

    Runs at IRQL <= DISPATCH_LEVEL.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER_LIST NetBufferList
        The NBL to reference.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if reference was acquired succesfully.
    NDIS_STATUS_RESOURCES if no resources are available to send the packet.
*/

static
NDIS_STATUS TXNblReference (
    IN  PNDIS_ADAPTER    pAdapter,
    IN  PNET_BUFFER_LIST NetBufferList )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

    if (pAdapter->ulDriverState & (DRIVER_STATE_PAUSE | DRIVER_STATE_PAUSED))
    {
        DBG_PRINT("paused when tx\n");
        return NDIS_STATUS_PAUSED;
    }
    if (pHardware->m_TxDescInfo.cnAvail < 2)
    {
        return NDIS_STATUS_RESOURCES;
    }

    NdisInterlockedIncrement(&pAdapter->nBusySend);
    NdisInterlockedIncrement(&SEND_REF_FROM_NBL(NetBufferList));
    return NDIS_STATUS_SUCCESS;
}  // TXNblReference


/*
    TXNblRelease

Description:
    This procedure releases a reference on a NBL that is being transmitted.
    If the last reference is released, the NBL is returned to the protocol.

    Runs at IRQL <= DISPATCH_LEVEL.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER_LIST NetBufferList
        The NBL to release.

    BOOLEAN fAtDispatch
        TRUE if the current IRQL is DISPATCH_LEVEL.

Return (None):
*/

static
VOID TXNblRelease (
    IN  PNDIS_ADAPTER    pAdapter,
    IN  PNET_BUFFER_LIST NetBufferList,
    IN  BOOLEAN          fAtDispatch )
{

    NdisInterlockedDecrement(&pAdapter->nBusySend);
    if (0 == NdisInterlockedDecrement(&SEND_REF_FROM_NBL(NetBufferList)))
    {
        NET_BUFFER_LIST_NEXT_NBL(NetBufferList) = NULL;

        NdisMSendNetBufferListsComplete(pAdapter->hAdapter, NetBufferList,
            fAtDispatch ? NDIS_SEND_COMPLETE_FLAGS_DISPATCH_LEVEL : 0);
    }
}  // TXNblRelease


/*
    SendPacket

Description:
    This function sends a single packet to hardware.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER pBuffer
        Pointer to NDIS net buffer.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise, an error code indicating
    failure.
*/

NDIS_STATUS SendPacket (
    PNDIS_ADAPTER pAdapter,
    PNET_BUFFER   pBuffer )
{
    PBUFFER_INFO pBufInfo = &pAdapter->TxBufInfo;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_TxDescInfo;
    NDIS_STATUS Status = NDIS_STATUS_FAILURE;
    ULONG FrameType;
    ULONG Length;
    int port = MAIN_PORT;

    pBufInfo->pCurrent = Alloc_DMA_Buffer(pBufInfo, pInfo->pCurrent);

    Length = CopyNetBuffer(pAdapter, pBuffer);
    if (pAdapter->ulUdpTestLimit)
        pAdapter->nCopySend++;
    if (pAdapter->nCopySend == pAdapter->ulUdpTestLimit)
        pBufInfo->pCurrent->pBuffer = pBuffer;

#if 0
    if ( pAdapter->ulTxChecksum )
        HardwareChecksum( pPacket, pHardware );
#endif
    HardwareSendPacket(pHardware);
    Status = NDIS_STATUS_SUCCESS;
    pHardware->m_cnCounter[port][OID_COUNTER_XMIT_OK]++;
    FrameType = FRAME_TYPE_FROM_SEND_NB(pBuffer);
    if (Length < 60)
        Length = 64;
    else
        Length += 4;
    if (NDIS_PACKET_TYPE_BROADCAST == FrameType)
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_BROADCAST_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_BROADCAST_BYTES_XMIT] += Length;
    }
    else if (NDIS_PACKET_TYPE_MULTICAST == FrameType)
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_MULTICAST_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_MULTICAST_BYTES_XMIT] += Length;
    }
    else
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_DIRECTED_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_DIRECTED_BYTES_XMIT] += Length;
    }
    if (pAdapter->ulUdpTestLimit &&
            pAdapter->nCopySend < pAdapter->ulUdpTestLimit)
        TXNblRelease(pAdapter, NBL_FROM_SEND_NB(pBuffer), TRUE);
    else
        pAdapter->nCopySend = 0;

    return Status;
}  // SendPacket


VOID MiniportProcessSGList (
    IN  PDEVICE_OBJECT       pDO,
    IN  PVOID                Reserved,
    IN  PSCATTER_GATHER_LIST pSGL,
    IN  PVOID                Context )
{
    PTCB pTcb = Context;
    PNDIS_ADAPTER pAdapter = pTcb->pAdapter;
    PTDesc pDesc;
    PTDesc pLastDesc;
    PTDesc pFirstDesc;
    PDMA_BUFFER pDma;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_TxDescInfo;
    NDIS_STATUS nsStatus = NDIS_STATUS_FAILURE;
    TDesc BufDesc;
    UINT Seg;
    int port = MAIN_PORT;
    ULONG FrameType;
    ULONG Length = pTcb->PacketLength;
    ULONG MdlOffset = NET_BUFFER_CURRENT_MDL_OFFSET(pTcb->NetBuffer);
    UINT32 Address;

#ifdef DBG
    if (pSGL->NumberOfElements != pTcb->BufferCount &&
            pSGL->NumberOfElements != pTcb->BufferCount + 1)
    {
        DbgPrint("buffer: %u %u\n", pSGL->NumberOfElements, pTcb->BufferCount);
    }
#endif
    pTcb->pSGL = pSGL;
    if ((int) pSGL->NumberOfElements > pHardware->m_TxDescInfo.cnAvail)
    {
        SendPacket(pAdapter, pTcb->NetBuffer);
        ReturnTcb(&pAdapter->TcbInfo, pTcb);
        return;
    }

    /* Remember the very first descriptor. */
    pFirstDesc = pInfo->pCurrent;
    pDesc = pFirstDesc;
    pLastDesc = pFirstDesc;

    pDma = Alloc_DMA_Buffer(&pAdapter->TxBufInfo, &BufDesc);

    for (Seg = 0; Seg < pSGL->NumberOfElements; Seg++)
    {

        /* Release this descriptor to hardware if not the very
           first one, not wanting the hardware to start
           transmitting this packet when descriptors are being set.
        */
        if (pLastDesc != pFirstDesc)
            ReleasePacket(pLastDesc);

        Address = NdisGetPhysicalAddressLow(pSGL->Elements[Seg].Address);
        Length = pSGL->Elements[Seg].Length;
        Address += MdlOffset;
        Length -= MdlOffset;
        MdlOffset = 0;
        SetTransmitBuffer(pDesc, Address);
        SetTransmitLength(pDesc, Length);

        /* Remember the last descriptor. */
        pLastDesc = pDesc;

        /* Allocate more descriptor. */
        GetTxPacket(pInfo, pDesc);

#ifdef SKIP_TX_INT
        ++pHardware->m_TxIntCnt;
#endif
    }

    /* If more than one descriptor is used for this buffer, move the
       buffer so that the mapping is freed only after all descriptors
       are transmitted.
    */
    pLastDesc->pReserved = BufDesc.pReserved;

    /* One more descriptor is allocated. */
    HardwareFreeTxPacket(pInfo);

    /* The last descriptor holds the packet so that it can be returned to
       network subsystem after all descriptors are transmitted.
    */
    pDma->Tcb = pTcb;
    pDma->pBuffer = pTcb->NetBuffer;

    /* pCurrent points to the last descriptor. */
    pInfo->pCurrent = pLastDesc;

    /* Release the first descriptor. */
    ReleasePacket(pFirstDesc);

#if 0
    if ( pAdapter->ulTxChecksum )
        HardwareChecksum( pPacket, pHardware );
#endif
    HardwareSendPacket(pHardware);
    pHardware->m_cnCounter[port][OID_COUNTER_XMIT_OK]++;
    FrameType = FRAME_TYPE_FROM_SEND_NB(pTcb->NetBuffer);
    Length = pTcb->PacketLength;
    if (Length < 60)
        Length = 64;
    else
        Length += 4;
    if (NDIS_PACKET_TYPE_BROADCAST == FrameType)
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_BROADCAST_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_BROADCAST_BYTES_XMIT] += Length;
    }
    else if (NDIS_PACKET_TYPE_MULTICAST == FrameType)
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_MULTICAST_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_MULTICAST_BYTES_XMIT] += Length;
    }
    else
    {
        pHardware->m_cnCounter[port]
            [OID_COUNTER_DIRECTED_FRAME_XMIT]++;
        pHardware->m_cnCounter[port]
            [OID_COUNTER_DIRECTED_BYTES_XMIT] += Length;
    }
}  // MiniportProcessSGList


/*
    SendBuffers

Description:
    This function sends a packet in several buffer descriptors to hardware.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER pBuffer
        Pointer to NDIS net buffer.

    ULONG PacketLength
        Length of the packet.

    ULONG BufferCount
        Number of buffers.

    BOOLEAN fAtDispatch
        Indicate dispatch level.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise, an error code indicating
    failure.
*/

NDIS_STATUS SendBuffers (
    IN  PNDIS_ADAPTER pAdapter,
    IN  PNET_BUFFER   pBuffer,
    IN  ULONG         PacketLength,
    IN  ULONG         BufferCount,
    IN  BOOLEAN       fAtDispatch )
{
    NDIS_STATUS Status;
    PSLIST_ENTRY pEntry;
    PTCB pTcb;
    PTCB_INFO pInfo = &pAdapter->TcbInfo;

    //
    // Get the next available TCB.
    //
    pEntry = NdisInterlockedPopEntrySList(&pInfo->listHead,
        &pInfo->listLock);
    pTcb = CONTAINING_RECORD(pEntry, TCB, listEntry);

    pTcb->NetBuffer = pBuffer;
    pTcb->PacketLength = PacketLength;
    pTcb->BufferCount = BufferCount;

    Status = NdisMAllocateNetBufferSGList(pInfo->hDma,
        pBuffer, pTcb, NDIS_SG_LIST_WRITE_TO_DEVICE,
        pTcb->ScatterGatherBuffer, pInfo->ScatterGatherSize);
    if (NDIS_STATUS_SUCCESS != Status)
    {
        PNET_BUFFER_LIST Nbl = NBL_FROM_SEND_NB(pBuffer);

        ReturnTcb(pInfo, pTcb);
        NET_BUFFER_LIST_STATUS(Nbl) = Status;
        TXNblRelease(pAdapter, Nbl, fAtDispatch);
    }

    return Status;
}  // SendBuffers

/* -------------------------------------------------------------------------- */

/*
    SendNetBuffer

Description:
    This function sends one packet in NDIS 6.0.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER NetBuffer
        Pointer to NDIS net buffer.

    BOOLEAN fAtDispatch
        Indicate dispatch level.

Return (NDIS_STATUS):
    NDIS_STATUS_PENDING if successful; otherwise, an error code indicating
    failure.
*/

NDIS_STATUS SendNetBuffer (
    IN  PNDIS_ADAPTER pAdapter,
    IN  PNET_BUFFER   NetBuffer,
    IN  BOOLEAN       fAtDispatch )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    ULONG BufferCount;
    ULONG PacketLength;
    int nCount;
    NDIS_STATUS Status = NDIS_STATUS_PENDING;

#ifdef DEBUG_COUNTER
    pHardware->m_nGood[COUNT_GOOD_SEND_PACKET]++;
#endif
    QueryPacket(pAdapter, NetBuffer, &BufferCount, &PacketLength);

    nCount = HardwareAllocPacket(pHardware, PacketLength, BufferCount);

    if (nCount > (int) BufferCount && BufferCount > 1)
    {
        Status = SendBuffers(pAdapter, NetBuffer, PacketLength, BufferCount,
            fAtDispatch);
    }
    else
    {
        Status = SendPacket(pAdapter, NetBuffer);
    }

    return Status;
}  // SendNetBuffer


/*
    TXTransmitQueuedSends

Description:
    This routine sends as many frames from the SendWaitList as it can.

    If there are not enough resources to send immediately, this function stops
    and leaves the remaining frames on the SendWaitList, to be sent once there
    are enough resources.

    Runs at IRQL <= DISPATCH_LEVEL

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    BOOLEAN fAtDispatch
        TRUE if the current IRQL is DISPATCH_LEVEL.

Return (None):
*/

VOID
TXTransmitQueuedSends (
    IN  PNDIS_ADAPTER pAdapter,
    IN  BOOLEAN       fAtDispatch )
{
    PHARDWARE pHardware = &pAdapter->Hardware;

    pAdapter->fSend = FALSE;

    // Enough transmit descriptors available to send packet.
    while (pHardware->m_TxDescInfo.cnAvail > 1)
    {
        NDIS_STATUS Status;
        PLIST_ENTRY pQueuedSend;
        PNET_BUFFER NetBuffer;

        //
        // Get the next NB that needs sending.
        //
        pQueuedSend = NdisInterlockedRemoveHeadList(&pAdapter->SendWaitList,
            &pAdapter->SendWaitListLock);
        if (!pQueuedSend)
        {
            //
            // There's nothing left that needs sending.  We're all done.
            //
            break;
        }

        NetBuffer = NB_FROM_SEND_WAIT_LIST(pQueuedSend);

        Status = SendNetBuffer(pAdapter, NetBuffer, fAtDispatch);
        if (NDIS_STATUS_SUCCESS != Status)
        {
             break;
        }
    }
}  // TXTransmitQueuedSends


/*
    TXFlushSendQueue

Description:
    This routine is called by the Halt or Reset handler to fail all
    the queued up Send NBLs because the device is either gone, being
    stopped for resource rebalance, or reset.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    NDIS_STATUS CompleteStatus
        The status code with which to complete each NBL.

Return (None):
*/

VOID TXFlushSendQueue (
    IN  PNDIS_ADAPTER pAdapter,
    IN  NDIS_STATUS   CompleteStatus )
{
    //
    // First, free anything queued in the driver.
    //

    while (TRUE)
    {
        PLIST_ENTRY pEntry;
        PNET_BUFFER NetBuffer;
        PNET_BUFFER_LIST NetBufferList;

        pEntry = NdisInterlockedRemoveHeadList(
            &pAdapter->SendWaitList, &pAdapter->SendWaitListLock);

        if (!pEntry)
        {
            // End of list -- nothing left to free.
            break;
        }

        NetBuffer = NB_FROM_SEND_WAIT_LIST(pEntry);
        NetBufferList = NBL_FROM_SEND_NB(NetBuffer);

        DEBUGP(MP_TRACE, ("[%p] Dropping Send NB: 0x%p.\n", pAdapter,
            NetBuffer));

        NET_BUFFER_LIST_STATUS(NetBufferList) = CompleteStatus;
        TXNblRelease(pAdapter, NetBufferList, FALSE);
    }
}  // TXFlushSendQueue

/* -------------------------------------------------------------------------- */

/*
    TransmitDone

Description:
    This routine cleans up the descriptors after transmission.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID TransmitDone (
    PNDIS_ADAPTER pAdapter )
{
    int iLast;
    TDescStat status;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_TxDescInfo;
    PTDesc pDesc;
    PDMA_BUFFER pDma;

    if (InterlockedIncrement(&pAdapter->TransmitAcquire) > 1)
    {
        goto transmit_done;
    }

    iLast = pInfo->iLast;

    while (pInfo->cnAvail < pInfo->cnAlloc)
    {
        /* Get next descriptor which is not hardware owned. */
        GetTransmittedPacket(pInfo, iLast, pDesc, status.ulData);
        if (status.tx.fHWOwned)
            break;

        pDma = DMA_BUFFER(pDesc);

        /* This descriptor contains a buffer. */
        if (pDma)
        {
            /* It is a mapped buffer. */
            if (pDma->Tcb)
            {
                pDma->pBuffer = pDma->Tcb->NetBuffer;
                ReturnTcb(&pAdapter->TcbInfo, pDma->Tcb);
                pDma->Tcb = NULL;
            }

            /* This descriptor contains the last buffer in the packet. */
            if (pDma->pBuffer)
            {
                /* Notify the network subsystem that the packet has been sent.
                */
                TXNblRelease(pAdapter, NBL_FROM_SEND_NB(pDma->pBuffer),
                    TRUE);

                /* Release the packet back to network subsystem. */
                pDma->pBuffer = NULL;
            }

            /* Free the transmitted buffer. */
            Free_DMA_Buffer(&pAdapter->TxBufInfo, pDma);
            pDesc->pReserved = NULL;
        }

        /* Free the transmitted descriptor. */
        FreeTransmittedPacket(pInfo, iLast);
    }
    pInfo->iLast = iLast;
    pAdapter->nNotSend = 0;

    if (pHardware->m_TxDescInfo.cnAvail >
            pHardware->m_TxDescInfo.cnAlloc / 4)
        TXTransmitQueuedSends(pAdapter, TRUE);

transmit_done:
    InterlockedDecrement(&pAdapter->TransmitAcquire);
}  // TransmitDone


/*
    TransmitReset

Description:
    This routine resets the descriptors to starting state after a software
    reset.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID TransmitReset (
    PNDIS_ADAPTER pAdapter )
{
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_TxDescInfo;
    int iLast;
    PTDesc pDesc;
    PDMA_BUFFER pDma;
    TDescStat status;

    iLast = pInfo->iLast;

    while (pInfo->cnAvail < pInfo->cnAlloc)
    {
        /* Get next descriptor which is hardware owned. */
        GetTransmittedPacket(pInfo, iLast, pDesc, status.ulData);
        if (status.tx.fHWOwned)
        {
            ReleaseDescriptor(pDesc, status);
        }

        pDma = DMA_BUFFER(pDesc);

        /* This descriptor contains a buffer. */
        if (pDma)
        {
            /* It is a mapped buffer. */
            if (pDma->Tcb)
            {
                pDma->pBuffer = pDma->Tcb->NetBuffer;
                ReturnTcb(&pAdapter->TcbInfo, pDma->Tcb);
                pDma->Tcb = NULL;
            }

            /* This descriptor contains the last buffer in the packet. */
            if (pDma->pBuffer)
            {
                /* Notify the network subsystem that the packet has not been
                   sent.
                */
                PNET_BUFFER_LIST Nbl = NBL_FROM_SEND_NB(pDma->pBuffer);

                NET_BUFFER_LIST_STATUS(Nbl) = NDIS_STATUS_FAILURE;
                TXNblRelease(pAdapter, Nbl, TRUE);

                /* Release the packet back to network subsystem. */
                pDma->pBuffer = NULL;
            }

            /* Free the buffer. */
            Free_DMA_Buffer(&pAdapter->TxBufInfo, pDma);
            pDesc->pReserved = NULL;
        }

        /* Free the descriptor. */
        FreeTransmittedPacket(pInfo, iLast);
    }
    pInfo->iLast = iLast;
    pAdapter->nNotSend = 0;

    TXFlushSendQueue(pAdapter, NDIS_STATUS_FAILURE);
}  // TransmitReset

/* -------------------------------------------------------------------------- */

/*
    TXQueueNetBufferForSend

Description:
    This routine inserts the NET_BUFFER into the SendWaitList, then calls
    TXTransmitQueuedSends to start sending data from the list.

    We use this indirect queue to send data because the miniport should try to
    send frames in the order in which the protocol gave them.  If we just sent
    the NET_BUFFER immediately, then it would be out-of-order with any data on
    the SendWaitList.

    Runs at IRQL <= DISPATCH_LEVEL

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    PNET_BUFFER NetBuffer
        NB to be transfered.

Return (None):
*/

VOID TXQueueNetBufferForSend (
    IN  PNDIS_ADAPTER pAdapter,
    IN  PNET_BUFFER   NetBuffer )
{
    NdisInterlockedIncrement(&pAdapter->nBusySend);

    //
    // Pin the original NBL with a reference, so it isn't completed until
    // we're done with its NB.
    //
    NdisInterlockedIncrement(&SEND_REF_FROM_NBL(NBL_FROM_SEND_NB(NetBuffer)));

    //
    // Insert the NB into the queue.  The caller will flush the queue when
    // it's done adding items to the queue.
    //
    NdisInterlockedInsertTailList(&pAdapter->SendWaitList,
        SEND_WAIT_LIST_FROM_NB(NetBuffer), &pAdapter->SendWaitListLock);
}  // TXQueueNetBufferForSend


/*
    MiniportCancelSend

Description:
    MiniportCancelSend cancels the transmission of all NET_BUFFER_LISTs that
    are marked with a specified cancellation identifier. Miniport drivers
    that queue send packets for more than one second should export this
    handler. When a protocol driver or intermediate driver calls the
    NdisCancelSendNetBufferLists function, NDIS calls the MiniportCancelSend
    function of the appropriate lower-level driver (miniport driver or
    intermediate driver) on the binding.

    Runs at IRQL <= DISPATCH_LEVEL.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PVOID CancelId
        All the packets with this Id should be cancelled.

Return (None):
*/

VOID MiniportCancelSend (
    IN  NDIS_HANDLE hAdapterContext,
    IN  PVOID       CancelId )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    UNREFERENCED_PARAMETER(CancelId);

    DEBUGP(MP_TRACE, ("[%p] ---> MPCancelSend\n", pAdapter));

    //
    // This miniport completes its sends quickly, so it isn't strictly
    // neccessary to implement MiniportCancelSend.
    //
    // If we did implement it, we'd have to walk the Adapter->SendWaitList
    // and look for any NB that points to a NBL where the CancelId matches
    // NDIS_GET_NET_BUFFER_LIST_CANCEL_ID(Nbl).  For any NB that so matches,
    // we'd remove the NB from the SendWaitList and set the NBL's status to
    // NDIS_STATUS_SEND_ABORTED, then complete the NBL.
    //

    DEBUGP(MP_TRACE, ("[%p] <--- MPCancelSend\n", pAdapter));
}


/*
    MiniportSendNetBufferLists

Description:
    This routine sends packets in NDIS 6.0.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNET_BUFFER_LIST NetBufferLists
        Pointer to net buffer lists containing NDIS packets.

    NDIS_PORT_NUMBER PortNumber
        Port number.

    ULONG SendFlags
        Indicate dispatch level.

Return (None):
*/

VOID MiniportSendNetBufferLists (
    IN  NDIS_HANDLE      hAdapterContext,
    IN  PNET_BUFFER_LIST NetBufferLists,
    IN  NDIS_PORT_NUMBER PortNumber,
    IN  ULONG            SendFlags )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PNET_BUFFER_LIST Nbl;
    PNET_BUFFER_LIST NextNbl = NULL;
    BOOLEAN fAtDispatch =
        (SendFlags & NDIS_SEND_FLAGS_DISPATCH_LEVEL) ? TRUE : FALSE;
    NDIS_STATUS Status;

    UNREFERENCED_PARAMETER(PortNumber);
    UNREFERENCED_PARAMETER(SendFlags);

    //
    // Each NET_BUFFER_LIST has a list of NET_BUFFERs.
    // Loop over all the NET_BUFFER_LISTs, sending each NET_BUFFER.
    //
    for (Nbl = NetBufferLists; Nbl != NULL; Nbl = NextNbl)
    {
        PNET_BUFFER NetBuffer;

        NextNbl = NET_BUFFER_LIST_NEXT_NBL(Nbl);

        //
        // Unlink the NBL and prepare our bookkeeping.
        //
        // We use a reference count to make sure that we don't send complete
        // the NBL until we're done reading each NB on the NBL.
        //
        NET_BUFFER_LIST_NEXT_NBL(Nbl) = NULL;
        SEND_REF_FROM_NBL(Nbl) = 0;

        Status = TXNblReference(pAdapter, Nbl);
        if (Status == NDIS_STATUS_SUCCESS)
        {
            NET_BUFFER_LIST_STATUS(Nbl) = NDIS_STATUS_SUCCESS;

            //
            // Queue each NB for transmission.
            //
            for (NetBuffer = NET_BUFFER_LIST_FIRST_NB(Nbl);
                    NetBuffer != NULL;
                    NetBuffer = NET_BUFFER_NEXT_NB(NetBuffer))
            {
                NBL_FROM_SEND_NB(NetBuffer) = Nbl;
                TXQueueNetBufferForSend(pAdapter, NetBuffer);
            }
            pAdapter->fSend = TRUE;

            TXNblRelease(pAdapter, Nbl, fAtDispatch);
        }
        else
        {
            //
            // We can't send this NBL now.  Indicate failure.
            //
            NET_BUFFER_LIST_STATUS(Nbl) = Status;

            NdisMSendNetBufferListsComplete(pAdapter->hAdapter, Nbl,
                fAtDispatch ? NDIS_SEND_COMPLETE_FLAGS_DISPATCH_LEVEL : 0);

            continue;
        }
    }

    //
    // Now actually go send each of the queued NBs.
    //
    if (!AcquireAdapter(pAdapter, FALSE))
        return;

    if (1 == InterlockedIncrement(&pAdapter->TransmitAcquire))
    {
        TXTransmitQueuedSends(pAdapter, fAtDispatch);
    }
    InterlockedDecrement(&pAdapter->TransmitAcquire);
    ReleaseAdapter(pAdapter);
}  // MiniportSendNetBufferLists
